"""
    This file is part of Accountant.

    Accountant is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Accountant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with Accountant.  If not, see <http://www.gnu.org/licenses/>.
"""
import arrow

from sqlalchemy import func, case, desc

from accountant import db


class Operation(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    pointed = db.Column(db.Boolean, nullable=False, default=False)
    operation_date = db.Column(db.Date, nullable=False)
    label = db.Column(db.String(500), nullable=False)
    value = db.Column(db.Numeric(15, 2), nullable=False)
    account_id = db.Column(db.Integer, db.ForeignKey('account.id'))
    scheduled_operation_id = db.Column(db.Integer,
                                       db.ForeignKey('scheduled_operation.id'))

    account = db.relationship('Account', backref=db.backref('operation',
                                                            lazy="dynamic"))

    category = db.Column(db.String(100), nullable=True)

    canceled = db.Column(db.Boolean, nullable=False)

    def __init__(self, pointed, label, value, account_id, operation_date=None,
                 category=None):
        self.pointed = pointed
        self.operation_date = operation_date
        self.label = label
        self.value = value
        self.account_id = account_id
        self.category = category

    @classmethod
    def get_for_account_and_month(cls, session, account, year, month):
        begin = arrow.get(year, month, 1)
        end = begin.ceil('month')

        return cls.get_for_account_and_range(session, account, begin.date(),
                                             end.date())

    @classmethod
    def get_for_account_and_range(cls, session, account, begin, end):
        if isinstance(account, int) or isinstance(account, str):
            account_id = account
        else:
            account_id = account.id

        base_query = session.query(
            cls,
            case(
                whens={cls.canceled: None},
                else_=func.sum(cls.value).over(
                    partition_by="canceled",
                    order_by=["operation_date", desc("value"), desc("label")])
            ).label("sold")
        ).filter(cls.account_id == account_id).order_by(
            desc(cls.operation_date),
            cls.value,
            cls.label,
        ).subquery()

        query = session.query(base_query).select_from(base_query)
        query = query.filter(base_query.c.operation_date >= str(begin),
                             base_query.c.operation_date <= str(end))

        return query

    @classmethod
    def get_account_status(cls, session, account, year, month):
        if isinstance(account, int) or isinstance(account, str):
            account_id = account
        else:
            account_id = account.id

        query = session.query(
            func.sum(case([(func.sign(cls.value) == -1, cls.value)],
                          else_=0)).label("expenses"),
            func.sum(case([(func.sign(cls.value) == 1, cls.value)],
                          else_=0)).label("revenues"),
            func.sum(cls.value).label("balance")
        ).filter(
            cls.account_id == account_id
        ).filter(
            func.date_trunc('month',
                            cls.operation_date) == "%s-%s-01" % (year, month)
        ).group_by(cls.account_id)

        return query

    @classmethod
    def get(cls, session, id):
        return session.query(cls).filter(cls.id == id).one()
